/*
   align_to_scf.c

   Bernd Senf - bsenf@fli-leibniz.de - FLI Jena e.V.

   2007-05-02 - initial revision
   2007-05-22 - set MAXLEN 4096 
   2007-12-11 - handle readnames with extensions like READ.1-5.fm16
   2008-05-01 - Only bzero the portion of SCORE[][] needed. This is a huge
                (~1000x) speedup. (James Bonfield)
              - Dynamically allocate SCORE[][] to remove MAXLEN issues.
   2009-01-29 - fixed memory bug with "no title"
 */

/*
#######################################################################
# This software has been created by FLI Jena e.V., Department of      #
# Genome Analysis (FLI).                                              #
# FLI hereby grants permission to use, copy, modify and distribute    #
# this software and its documentation for non-commercial purposes     #
# without fee at the users own risk on the basis set out below.       #
#                                                                     #
# FLI neither undertakes nor accepts any duty whether contractual or  #
# otherwise in connection with the software, its use or the use of    #
# any derivative, and makes no representations or warranties, express #
# or implied, concerning the software, its suitability, fitness for   #
# a particular purpose or non-infringement.                           #
# In no event shall the authors of the software or FLI be responsible #
# or liable for any loss or damage whatsoever arising in any way      #
# directly or indirectly out of the use of this software or its       #
# derivatives, even if advised of the possibility of such damage.     #
#                                                                     #
# Our software can be freely distributed under the conditions set out #
# above, and must contain this copyright notice.                      #
#######################################################################
*/

#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include <ctype.h>			/* isalnum, isspace */
#include <string.h> 		/* strlen */
#include <strings.h>  		/* bzero */

const char *VERSION="1.06";

#define max(A,B) (((A) > (B) ) ? (A): (B))
#define min(A,B) (((A) < (B) ) ? (A): (B))

#define MAXGAPS 1024		/* max count of gaps for long stretched sequence */

#ifndef MAXLEN
#define MAXLEN 100000
#endif

#define DYNAMIC

#ifdef DYNAMIC
int *SCORE_x;
int SCORE_cols;
/* Also allows for provision for SCORE matrix being sparse in the future */
#define SCORE(i,j) (SCORE_x[(i)*SCORE_cols+(j)])
#else
int SCORE_x[MAXLEN][MAXLEN];
#define SCORE(i,j) (SCORE_x[(i)][(j)])
#endif

int ialn;

struct 
{
	int iace;
	int itrace;
	int iscore;
} ALIGN[MAXGAPS];

struct {
	int itrace; 		/* trace index */
	int iscore; 		/* score value */
	int n;				/* acount */
} MAXSCORE[MAXLEN]; 	/* ace index */


void show_aligns(char *title, FILE *FpOUT)
{
	if (ialn > 0 ) 
	{
		int i;
		fprintf(FpOUT,">%s\n",title);
		i=ialn;
		while (i--) 
		{
			int iace;
			int iscore;
			int itrace;
			
			iace=ALIGN[i].iace;
			itrace=ALIGN[i].itrace;
			iscore=ALIGN[i].iscore;
			fprintf(FpOUT,"Align_to_SCF %d %d %d %d\n",iace-iscore+2,iace+1,itrace-iscore+2,itrace+1);
		}
		fprintf(FpOUT,"\n");
		fflush(stdout);
	} 
	else 
	{
		fprintf(stdout,">%s - NOALIGNS\n",title); /* fatal - only occure if ace[aaaaa] => trace[ttttt] */
		fflush(stdout);
	}
}


void store_aligns(int lace, int ltrace, int itol)
{
	int iace=0;
	int itrace=0;
	int iscore=0;
	int i;
	int j;
	
	for (i=lace-1;i>0 && iscore==0;i--) 
	{
		for (j=ltrace-1;j>=max(0,ltrace-itol);j--) 
		{
			if ((SCORE(i,j)) > iscore )
			{
				iace=i;
				itrace=j;
				iscore=SCORE(i,j);
			}
		}
	}
	if (iace==0 || iscore==0 || itrace==0 ) return; /* Finished */
	
	ALIGN[ialn].iace=iace;
	ALIGN[ialn].itrace=itrace;
	ALIGN[ialn].iscore=iscore;
	
	ialn++;
	if (iace && itrace ) store_aligns(iace-iscore+1,itrace-iscore+1,5);
	return;	
}

void store_maxscore(int i,int j,int n)
{
	SCORE(i,j)=n;
	
	if (MAXSCORE[i].iscore == n ) 
	{
		MAXSCORE[i].iscore=n;
		MAXSCORE[i].itrace=j;
		MAXSCORE[i].n++;
	}
	
	if (MAXSCORE[i].iscore < n ) 
	{
		MAXSCORE[i].iscore=n;
		MAXSCORE[i].itrace=j;
		MAXSCORE[i].n=1;
	}
}

void find_aligns(char *title, char *ace, char *trace, int lenace, int lentrace)
{
	int i,j;
	int ngaps;
	int maxdash;
	char *cp;
	
	int lastj;  /* last good ace[i] */
	int lasts;	/* last good score from ace[i] */
	
	int nscore; /* max score pro line */
	int nidxs;	/* count repeats max score ( 1 - single hit, 2 - two possible repeats, ... ) */
	 
	if (lenace > MAXLEN)
	    lenace = MAXLEN;
	if (lentrace > MAXLEN)
	    lentrace = MAXLEN;

#ifdef DYNAMIC
	SCORE_x = (int *)calloc(lenace*lentrace, sizeof(int));
	SCORE_cols = lentrace;
#else
	for (i = 0; i < lenace; i++)
	    for (j = 0; j < lentrace; j++)
		SCORE(i,j) = 0;
#endif
	bzero(MAXSCORE,lenace * sizeof(*MAXSCORE));
	
	nscore=nidxs=lastj=lasts=ngaps=0;
	
	cp=ace;
	maxdash=0;
	for (cp=ace;*cp;cp++) 
	{
		if (*cp=='-') maxdash++;
	}
	
	/* now FILL SCORE MATRIX */
	
	for (i=0;i<lenace;i++) 
	{
		int diff;
		int jmin;
		int jmax;
		
		if ( ace[i]=='-') 
		{
			int j;
			ngaps++;
			if (i > 0 ) 
			{	// really - I found ace sequence beginning with '-atcg....' !!!
				j=MAXSCORE[i-1].itrace;
				MAXSCORE[i].itrace=j;
				MAXSCORE[i].n=MAXSCORE[i-1].n;
			}		
			continue;
		}
		diff=lentrace-(lenace-maxdash)+4;
		
		
		if ( i>0 && MAXSCORE[i-1].n==1 )
		{
			jmin=MAXSCORE[i-1].itrace+1;		/* next pos */
			jmax=min(jmin+2,lentrace);
		}
		else
		{
			jmin=max(0,i-ngaps);
			jmax=min(i+diff,lentrace);
		}

		for (j=jmin;j<jmax;j++) 
		{
			if (ace[i]==trace[j]) {
				int n;
								
				n=1;			
				store_maxscore(i,j,n);	
				
				if (i && j && (n=SCORE(i-1,j-1))>0) /* second hit */
				{
					int k;
					
					n++;
					store_maxscore(i,j,n);		     /* n = 2 */
					
					if (MAXSCORE[i-1].n == 1 ) 
					{
						int kk; 					 /* n == 2,3,4,... */
						 
						k=1;
					
						while (i+k < lenace && ace[i+k]==trace[j+k] )
						{
							n++;
							k++;
						}
						k--;
						
						if (k) 
						{					        /* one or more k-linear hits */
							kk=k;					/* store old k; */
							
							while (k > 0) 
							{
								store_maxscore(i+k,j+k,n--);	
								k--;
							}
							i=i+kk;						
							j=lentrace;				/* do not search more then one times */
						}
					}						
				}
			}	
		}
	}
	
	/* prepare Show Aligns */
	ialn=0;
	store_aligns(lenace,lentrace,lentrace);
	
#ifdef DYNAMIC
	free(SCORE_x);
#endif

	return;
}


#define MAXLINE (20*1024)

void do_parse_file(char *fname, char *fout, FILE *FpIN, FILE *FpOUT)
{
	char *line;
	char *title;
	char *fascf;
	char *faphd;
	
	static int i,j,n;					/* number of fasta sequence within this file */
	int lace;
	int ltrace;
	
	if ( (line=(char*)calloc(MAXLINE,1))==NULL) { fprintf(stderr,"malloc error (line)\n");   return; }
	if ((title=(char*)calloc(MAXLINE,1))==NULL) { fprintf(stderr,"malloc error (title)\n");  return; }
	if ((fascf=(char*)calloc(MAXLINE,1))==NULL) { fprintf(stderr,"malloc error (fa-scf)\n"); return; }
	if ((faphd=(char*)calloc(MAXLINE,1))==NULL) { fprintf(stderr,"malloc error (fa-phd)\n"); return; }
	
	
	n=0;
	while(fgets(line,MAXLINE,FpIN)!=NULL) 
	{
		/* three lines each jobs 
		   > NAME
		   fasta_seq_from_ace
		   fasta_seq_from_trace_or_phd
		*/   
		if (line[0]!='>' ) 
		{ 
			fprintf(stderr,"Error in %s at line %d (no title):\n%s\n",fname,n,line); 
			return; 
		}
		
		i=1;
		while (i < (MAXLINE-2) && isspace(line[i] )) i++;
		
		j=0;
		while (i < (MAXLINE-1) && !isspace(line[i]) && j < MAXLINE-1 )
		{
			title[j++]=line[i++];
		}
		title[j]='\0';
		
		n++;
		if (fgets(fascf,MAXLINE,FpIN)==NULL || fascf[0]=='>') 
		{
			fprintf(stderr,"Error in %s at line %d (no sequence ATGC):\n%s\n",fname,n,fascf); 
			return;
		}
		lace=strlen(fascf);
		while(lace>1 && isspace(fascf[lace-1])) { fascf[lace-1]='\0'; lace--; }
		
		
		n++;
		if (fgets(faphd,MAXLINE,FpIN)==NULL || faphd[0]=='>' ) 
		{
			fprintf(stderr,"Error in %s at line %d (no sequence ATGC):\n%s\n",fname,n,faphd); 
			return;
		}
		ltrace=strlen(faphd);
		while(ltrace>1 && isspace(faphd[ltrace-1])) { faphd[ltrace-1]='\0'; ltrace--; }
		
		if (lace-ltrace > MAXGAPS-10 ) {
			fprintf(stderr,"\n");
			fprintf(stderr,"(%s) to obscure;\n",title);
			fprintf(stderr,"RE-Compile align_to_scf with larger MAXGAPS (%d)\n",MAXGAPS);
			fprintf(stderr,"\n");
		}
		else	
		{
			find_aligns(title,fascf,faphd,lace,ltrace);
			show_aligns(title,FpOUT);
		}		
	}
	
	/* be nice */
	if (line  != NULL) { free(line);  line=NULL;  }
	if (title != NULL) { free(title); title=NULL; }
	if (fascf != NULL) { free(fascf); fascf=NULL; }
	if (faphd != NULL) { free(faphd); faphd=NULL; }
}

void usage(void) {
    fprintf(stderr,
			"\n"
			"VERSION align_to_scf %s\n"
			"\n"
            "Usage:\n"
            "   align_to_scf [-h] [-V] -f file.aln -o file.qry\n"
			"\n"
            "Example(s):\n\n"
			"   align_to_scf -f 454Contigs.qry -o 454Contigs.aln\n\n"
			"\n",VERSION);
    exit(1);
}

int main(int argc, char **argv)
{
	int c;
	
	char *input;
	char *output;

	FILE *FpIN;
	FILE *FpOUT;
	
	FpIN=FpOUT=NULL;
	input=output=NULL;
	
    while ((c = getopt(argc, argv, "Vhf:o:")) != -1) 
	{
        switch (c) 
		{
        	case 'f':
            	input = optarg;
            	break;
        	case 'o':
            	output = optarg;
            	break;
			case 'h':
			case 'V':
         	default:
            	usage();
        }
    }

	if ( !input || !output ) usage();
		
	if ((FpIN=fopen(input,"r")) != NULL )
	{
		if ((FpOUT=fopen(output,"w")) != NULL ) {
			do_parse_file(input,output,FpIN,FpOUT);
		} else {
			perror(output);
		}
	} else {
		perror(input);
	}
	
	if (FpIN)  fclose (FpIN);
	if (FpOUT) fclose (FpOUT);
	
	exit(0);
}
